@@ -1,6 +1,7 @@
 /* connection_type.c - python interface to connection objects
  *
  * Copyright (C) 2003-2019 Federico Di Gregorio <fog@debian.org>
+ * Copyright (C) 2020-2021 The Psycopg Team
  *
  * This file is part of psycopg.
  *
@@ -405,10 +406,22 @@ psyco_conn_tpc_recover(connectionObject *self, PyObject *dummy)
 static PyObject *
 psyco_conn_enter(connectionObject *self, PyObject *dummy)
 {
+    PyObject *rv = NULL;
+
     EXC_IF_CONN_CLOSED(self);
 
+    if (self->entered) {
+        PyErr_SetString(ProgrammingError,
+            "the connection cannot be re-entered recursively");
+        goto exit;
+    }
+
+    self->entered = 1;
     Py_INCREF(self);
-    return (PyObject *)self;
+    rv = (PyObject *)self;
+
+exit:
+    return rv;
 }
 
 
@@ -426,6 +439,9 @@ psyco_conn_exit(connectionObject *self, PyObject *args)
         goto exit;
     }
 
+    /* even if there will be an error, consider ourselves out */
+    self->entered = 0;
+
     if (type == Py_None) {
         if (!(tmp = PyObject_CallMethod((PyObject *)self, "commit", NULL))) {
             goto exit;
@@ -968,7 +984,7 @@ psyco_conn_lobject(connectionObject *self, PyObject *args, PyObject *keywds)
     Dprintf("psyco_conn_lobject: new lobject for connection at %p", self);
     Dprintf("psyco_conn_lobject:     parameters: oid = %u, mode = %s",
             oid, smode);
-    Dprintf("psyco_conn_lobject:     parameters: new_oid = %d, new_file = %s",
+    Dprintf("psyco_conn_lobject:     parameters: new_oid = %u, new_file = %s",
             new_oid, new_file);
 
     if (new_file)
@@ -1297,55 +1313,6 @@ static struct PyGetSetDef connectionObject_getsets[] = {
 
 /* initialization and finalization methods */
 
-RAISES_NEG static int
-obscure_password(connectionObject *conn)
-{
-    PQconninfoOption *options;
-    PyObject *d = NULL, *v = NULL, *dsn = NULL;
-    char *tmp;
-    int rv = -1;
-
-    if (!conn || !conn->dsn) {
-        return 0;
-    }
-
-    if (!(options = PQconninfoParse(conn->dsn, NULL))) {
-        /* unlikely: the dsn was already tested valid */
-        return 0;
-    }
-
-    if (!(d = psyco_dict_from_conninfo_options(
-            options, /* include_password = */ 1))) {
-        goto exit;
-    }
-    if (NULL == PyDict_GetItemString(d, "password")) {
-        /* the dsn doesn't have a password */
-        rv = 0;
-        goto exit;
-    }
-
-    /* scrub the password and put back the connection string together */
-    if (!(v = Text_FromUTF8("xxx"))) { goto exit; }
-    if (0 > PyDict_SetItemString(d, "password", v)) { goto exit; }
-    if (!(dsn = psyco_make_dsn(Py_None, d))) { goto exit; }
-    if (!(dsn = psyco_ensure_bytes(dsn))) { goto exit; }
-
-    /* Replace the connection string on the connection object */
-    tmp = conn->dsn;
-    psyco_strdup(&conn->dsn, Bytes_AS_STRING(dsn), -1);
-    PyMem_Free(tmp);
-
-    rv = 0;
-
-exit:
-    PQconninfoFree(options);
-    Py_XDECREF(v);
-    Py_XDECREF(d);
-    Py_XDECREF(dsn);
-
-    return rv;
-}
-
 static int
 connection_setup(connectionObject *self, const char *dsn, long int async)
 {
@@ -1356,7 +1323,7 @@ connection_setup(connectionObject *self, const char *dsn, long int async)
             self, async, Py_REFCNT(self)
       );
 
-    if (0 > psyco_strdup(&self->dsn, dsn, -1)) { goto exit; }
+    if (!(self->dsn = conn_obscure_password(dsn))) { goto exit; }
     if (!(self->notice_list = PyList_New(0))) { goto exit; }
     if (!(self->notifies = PyList_New(0))) { goto exit; }
     self->async = async;
@@ -1378,7 +1345,7 @@ connection_setup(connectionObject *self, const char *dsn, long int async)
         goto exit;
     }
 
-    if (conn_connect(self, async) != 0) {
+    if (conn_connect(self, dsn, async) != 0) {
         Dprintf("connection_init: FAILED");
         goto exit;
     }
@@ -1390,13 +1357,6 @@ connection_setup(connectionObject *self, const char *dsn, long int async)
         self, Py_REFCNT(self));
 
 exit:
-    /* here we obfuscate the password even if there was a connection error */
-    {
-        PyObject *ptype = NULL, *pvalue = NULL, *ptb = NULL;
-        PyErr_Fetch(&ptype, &pvalue, &ptb);
-        obscure_password(self);
-        PyErr_Restore(ptype, pvalue, ptb);
-    }
     return rv;
 }
 
